from ast import *
from typing import Optional

from ..util import CompilingNodeTransformer

"""
Checks that there was an import of dataclass if there are any class definitions
"""


ALLOWED_TYPING_IMPORTS = {"Dict", "List", "Union", "Self"}


class RewriteImportTyping(CompilingNodeTransformer):
    step = "Checking import and usage of typing"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.imports = set()

    def visit_ImportFrom(self, node: ImportFrom) -> Optional[ImportFrom]:
        if node.module != "typing":
            return node

        for n in node.names:
            if n.name not in ALLOWED_TYPING_IMPORTS:
                raise ValueError(
                    f"Only the following imports from typing are allowed: {ALLOWED_TYPING_IMPORTS}, found {n.name}"
                )
            if n.asname is not None:
                raise ValueError("Imports from typing cannot be aliased")
            self.imports.add(n.name)
        return None

    def visit_Name(self, node: Name) -> Optional[Name]:
        if node.id in ALLOWED_TYPING_IMPORTS and node.id not in self.imports:
            raise ValueError(
                f"{node.id} used, which is a keyword for special OpShin types, but typing not imported. Please add 'from typing import {node.id}'"
            )
        return node

    def visit_ClassDef(self, node: ClassDef) -> ClassDef:
        for i, attribute in enumerate(node.body):
            if isinstance(attribute, FunctionDef):
                for j, arg in enumerate(attribute.args.args):
                    if isinstance(arg.annotation, Name) and arg.annotation.id == "Self":
                        assert (
                            "Self" in self.imports
                        ), "Self used but not imported from typing. Please add 'from typing import Self'"
                        node.body[i].args.args[j].annotation.idSelf = node.name
                    if (
                        isinstance(arg.annotation, Subscript)
                        and arg.annotation.value.id == "Union"
                    ):
                        assert (
                            "Union" in self.imports
                        ), "Union used but not imported from typing. Please add 'from typing import Union'"
                        for k, s in enumerate(arg.annotation.slice.elts):
                            if isinstance(s, Name) and s.id == "Self":
                                assert (
                                    "Self" in self.imports
                                ), "Self used but not imported from typing. Please add 'from typing import Self'"
                                node.body[i].args.args[j].annotation.slice.elts[
                                    k
                                ].idSelf = node.name

                if (
                    isinstance(attribute.returns, Name)
                    and attribute.returns.id == "Self"
                ):
                    assert (
                        "Self" in self.imports
                    ), "Self used but not imported from typing. Please add 'from typing import Self'"
                    node.body[i].returns.idSelf = node.name

        return node



"""Rewrite rule to prevent asserting None values."""

import ast
from typing import Optional, cast

from .rewrite_cast_condition import SPECIAL_BOOL
from ..typed_ast import TypedAssert
from ..type_impls import UnitInstanceType, InstanceType, UnitType
from ..util import CompilingNodeTransformer


class RewriteAssertNone(CompilingNodeTransformer):
    """Prevent asserting None values as this would always fail."""

    step = "Checking for assertions of None values"

    def visit_Assert(self, node: TypedAssert) -> Optional[ast.AST]:
        """Prevent asserting None values as this would always fail."""
        # Check if the test expression is a call that returns None
        if (
            isinstance(node.test, ast.Call)
            and hasattr(node.test, "typ")
            and (
                node.test.typ == UnitInstanceType
                or (
                    isinstance(node.test.typ, InstanceType)
                    and isinstance(node.test.typ.typ, UnitType)
                )
            )
        ):
            raise SyntaxError(
                f"Asserting a function call that returns None at line {node.lineno}. "
                "This would always fail as it's equivalent to 'assert False'. "
                "The function likely performs its own assertions internally."
            )

        # Check if this is a bool() call wrapping a function that returns None
        # This handles the case where RewriteConditions has already wrapped the call
        if (
            isinstance(node.test, ast.Call)
            and isinstance(node.test.func, ast.Name)
            and node.test.func.id.startswith(
                SPECIAL_BOOL
            )  # This is the special bool from RewriteConditions
            and len(node.test.args) == 1
            and isinstance(node.test.args[0], ast.Call)
            and hasattr(node.test.args[0], "typ")
            and (
                node.test.args[0].typ == UnitInstanceType
                or (
                    isinstance(node.test.args[0].typ, InstanceType)
                    and isinstance(node.test.args[0].typ.typ, UnitType)
                )
            )
        ):
            # Get the original function name for a better error message
            func_name = "<function>"
            if isinstance(node.test.args[0].func, ast.Name) and hasattr(
                node.test.args[0].func, "id"
            ):
                func_name = node.test.args[0].func.id

            raise SyntaxError(
                f"Asserting a function call that returns None at line {node.lineno}. "
                f"Function '{func_name}' returns None, so 'assert {func_name}(...)' "
                "would always fail as it's equivalent to 'assert False'. "
                "The function likely performs its own assertions internally."
            )

        return node



"""
The PlutusV3 ledger API.
All classes involved in defining the ScriptContext passed by the node.
"""

from dataclasses import dataclass
from typing import Dict, List, Union

from pycardano import Datum as Anything, PlutusData
from opshin.std.fractions import *


@dataclass(unsafe_hash=True)
class NoScriptHash(PlutusData):
    """
    Indicates that there is no script associated with an output
    """

    CONSTR_ID = 1


# A transaction id, a 64 bytes long hash of the transaction body (also called transaction hash).
#
# Example value: bytes.fromhex("842a4d37b036da6ab3c04331240e67d81746beb44f23ad79703e026705361956")
TxId = bytes


@dataclass(unsafe_hash=True)
class TrueData(PlutusData):
    """
    A Datum that represents True in Haskell implementations.
    It is thus used as an encoding for True in the ScriptContext.

    Example value: TrueData()
    """

    CONSTR_ID = 1


@dataclass(unsafe_hash=True)
class FalseData(PlutusData):
    """
    A Datum that represents False in Haskell implementations.
    It is thus used as an encoding for False in the ScriptContext.

    Example value: FalseData()
    """

    CONSTR_ID = 0


# A Datum that represents a boolean value in Haskell implementations.
# It is thus used as an encoding for booleans in the ScriptContext.
#
# Example value: TrueData()
BoolData = Union[TrueData, FalseData]


@dataclass(unsafe_hash=True)
class BoxedInt(PlutusData):
    """
    A boxed integer, used to represent an optional integer value
    """

    CONSTR_ID = 0
    value: int


@dataclass(unsafe_hash=True)
class NoValue(PlutusData):
    """
    An empty value (None case of Optional / Maybe)
    """

    CONSTR_ID = 1


Lovelace = int
OptionalInt = Union[BoxedInt, NoValue]
OptionalLovelace = OptionalInt


@dataclass(unsafe_hash=True)
class TxOutRef(PlutusData):
    """
    A reference to a transaction output (hash/id + index)
    """

    CONSTR_ID = 0

    id: TxId
    idx: int


# A public key hash, used to identify signatures provided by a wallet
# in aiken, this is called VerificationKeyHash
PubKeyHash = bytes


@dataclass(unsafe_hash=True)
class PubKeyCredential(PlutusData):
    """
    Part of an address that is authenticated by a public key hash

    Example value: PubKeyCredential(bytes.fromhex("c06ddaad12fc4ded18e56feac72957c1aa75fce6096b40e63ec88274"))
    """

    CONSTR_ID = 0
    credential_hash: PubKeyHash


# A validator hash, used to identify signatures provided by a smart contract
ValidatorHash = bytes


@dataclass(unsafe_hash=True)
class ScriptCredential(PlutusData):
    """
    Part of an address that is authenticated by a smart cotnract

    Example value: ScriptCredential(bytes.fromhex("c06ddaad12fc4ded18e56feac72957c1aa75fce6096b40e63ec88274"))
    """

    CONSTR_ID = 1
    credential_hash: ValidatorHash


# A credential, either smart contract or public key hash
Credential = Union[PubKeyCredential, ScriptCredential]


@dataclass(unsafe_hash=True)
class StakingHash(PlutusData):
    """
    Indicates that the stake of this address is controlled by the associated credential
    """

    CONSTR_ID = 0
    value: Credential


@dataclass(unsafe_hash=True)
class StakingPtr(PlutusData):
    """
    Indicates that the stake of this address is controlled by the associated pointer.

    In an address, a chain pointer refers to a point of the chain containing a stake key registration certificate.
    A point is identified by the 3 coordinates in this object.
    """

    CONSTR_ID = 1
    # an absolute slot number
    slot_no: int
    # a transaction index (within that slot)
    tx_index: int
    # a (delegation) certificate index (within that transaction)
    cert_index: int


# Part of an address that controls who can delegate the stake associated with an address
StakingCredential = Union[StakingHash, StakingPtr]


@dataclass(unsafe_hash=True)
class NoStakingCredential(PlutusData):
    """
    Indicates that this address has no staking credentials.
    Its funds can not be delegated.
    """

    CONSTR_ID = 1


@dataclass(unsafe_hash=True)
class SomeStakingCredential(PlutusData):
    """
    Indicates that this address has staking credentials.
    Its funds can be delegated by the credentialed user.
    """

    CONSTR_ID = 0
    staking_credential: StakingCredential


@dataclass(unsafe_hash=True)
class Address(PlutusData):
    """
    A Shelley address, consisting of a payment and staking credential
    """

    CONSTR_ID = 0

    payment_credential: Credential
    staking_credential: Union[NoStakingCredential, SomeStakingCredential]


# The policy Id of a token
PolicyId = bytes

# The name of a token in bytes (not textual representation!)
TokenName = bytes

# The Plutus representation of amounts of tokens being spent, sent or minted
# It is a two-layered dictionary that stores for each policy id and token name
# the amount of the token that is being sent/minted/burned etc
#
# Lovelace is represented with policy id b"" and token name b""
Value = Dict[PolicyId, Dict[TokenName, int]]

# A hash of a Datum
DatumHash = bytes


@dataclass(unsafe_hash=True)
class SomeDatumHash(PlutusData):
    """
    Indicates that there is a datum associated with this output, which has the given hash.
    """

    CONSTR_ID = 1
    datum_hash: DatumHash


@dataclass(unsafe_hash=True)
class SomeScriptHash(PlutusData):
    """
    Indicates that there is a script associated with this output, which has the given hash.
    """

    CONSTR_ID = 0
    script_hash: DatumHash


MaybeScriptHash = Union[SomeScriptHash, NoScriptHash]

# The abstract super type of any object in opshin.
# Use if you don't know what kind of object is being passed or if it doesn't matter.
BuiltinData = Anything


# An abstract type annotation that something is supposed to be used as a redeemer.
Redeemer = BuiltinData


# An abstract type annotation that something is supposed to be used as a datum.
Datum = BuiltinData


@dataclass(unsafe_hash=True)
class NoOutputDatum(PlutusData):
    """
    Indicates that there is no datum associated with an output
    """

    CONSTR_ID = 0


@dataclass(unsafe_hash=True)
class SomeOutputDatumHash(PlutusData):
    """
    Indicates that there is an datum associated with an output, which has the attached hash
    """

    CONSTR_ID = 1
    datum_hash: DatumHash


@dataclass(unsafe_hash=True)
class SomeOutputDatum(PlutusData):
    """
    Indicates that there is an datum associated with an output, which is inlined and equal to the attached datum
    """

    CONSTR_ID = 2
    datum: Datum


# Possible cases of datum association with an output
OutputDatum = Union[NoOutputDatum, SomeOutputDatumHash, SomeOutputDatum]


@dataclass(unsafe_hash=True)
class NoScriptHash(PlutusData):
    """
    Indicates that there is no script associated with an output
    """

    CONSTR_ID = 1


@dataclass(unsafe_hash=True)
class TxOut(PlutusData):
    """
    The plutus representation of an transaction output, consisting of
    - address: address owning this output
    - value: tokens associated with this output
    - datum: datum associated with this output
    - reference_script: reference script associated with this output
    """

    CONSTR_ID = 0

    address: Address
    value: Value
    datum: OutputDatum
    reference_script: Union[NoScriptHash, SomeScriptHash]


@dataclass(unsafe_hash=True)
class TxInInfo(PlutusData):
    """
    The plutus representation of an transaction output, that is consumed by the transaction.
    """

    CONSTR_ID = 0

    out_ref: TxOutRef
    resolved: TxOut


@dataclass(unsafe_hash=True)
class DCertDelegRegKey(PlutusData):
    CONSTR_ID = 0
    value: StakingCredential


@dataclass(unsafe_hash=True)
class DCertDelegDeRegKey(PlutusData):
    CONSTR_ID = 1
    value: StakingCredential


@dataclass(unsafe_hash=True)
class DCertDelegDelegate(PlutusData):
    CONSTR_ID = 2
    delegator: StakingCredential
    delegatee: PubKeyHash


@dataclass(unsafe_hash=True)
class DCertPoolRegister(PlutusData):
    CONSTR_ID = 3
    pool_id: PubKeyHash
    pool_vfr: PubKeyHash


@dataclass(unsafe_hash=True)
class DCertPoolRetire(PlutusData):
    CONSTR_ID = 4
    retirement_certificate: PubKeyHash
    epoch: int


@dataclass(unsafe_hash=True)
class DCertGenesis(PlutusData):
    CONSTR_ID = 5


@dataclass(unsafe_hash=True)
class DCertMir(PlutusData):
    CONSTR_ID = 6


Certificate = Union[
    DCertDelegRegKey,
    DCertDelegDeRegKey,
    DCertDelegDelegate,
    DCertPoolRegister,
    DCertPoolRetire,
    DCertGenesis,
    DCertMir,
]


# POSIX time in milliseconds, typically originating from the Cardano node.
# Note that Python's `time.time()` provides seconds, so remember to multiply by 1000.
POSIXTime = int


@dataclass(unsafe_hash=True)
class NegInfPOSIXTime(PlutusData):
    """
    Negative infinite POSIX time, used to indicate that there is no lower bound for the execution of this transaction
    """

    CONSTR_ID = 0


@dataclass(unsafe_hash=True)
class FinitePOSIXTime(PlutusData):
    """
    Finite POSIX time, used to indicate that there is a lower or upper bound for the execution of this transaction
    Note: POSIX time is in milliseconds, while Python's time.time() provides seconds, so remember to multiply/divide by 1000 when interacting with off-chain code.
    """

    CONSTR_ID = 1
    time: POSIXTime


@dataclass(unsafe_hash=True)
class PosInfPOSIXTime(PlutusData):
    """
    Infinite POSIX time, used to indicate that there is no upper bound for the execution of this transaction
    """

    CONSTR_ID = 2


ExtendedPOSIXTime = Union[NegInfPOSIXTime, FinitePOSIXTime, PosInfPOSIXTime]


@dataclass(unsafe_hash=True)
class UpperBoundPOSIXTime(PlutusData):
    """
    Upper bound for the execution of this transaction
    """

    CONSTR_ID = 0
    limit: ExtendedPOSIXTime
    closed: BoolData


@dataclass(unsafe_hash=True)
class LowerBoundPOSIXTime(PlutusData):
    """
    Lower bound for the execution of this transaction
    """

    CONSTR_ID = 0
    limit: ExtendedPOSIXTime
    closed: BoolData


@dataclass(unsafe_hash=True)
class POSIXTimeRange(PlutusData):
    """
    Time range in which this transaction can be executed
    Note: The transaction can be submitted at any point of time in the range.
    Deriving the actual submission time can be tricky. Best practices use only the upper or lower bound, depending on the use case.
    For example, if you accrue interest over time, use only the upper bound as the submission time, which is guaranteed to be after the actual submission time, but can be set tightly by the user to match the actual submission time.
    """

    CONSTR_ID = 0

    lower_bound: LowerBoundPOSIXTime
    upper_bound: UpperBoundPOSIXTime


@dataclass(unsafe_hash=True)
class Minting(PlutusData):
    """
    Script purpose indicating that the given policy id is being minted or burned
    """

    CONSTR_ID = 0
    policy_id: PolicyId


@dataclass(unsafe_hash=True)
class Spending(PlutusData):
    """
    Script purpose indicating that the given transaction output is being spent, which is
    owned by the invoked contract
    """

    CONSTR_ID = 1
    tx_out_ref: TxOutRef


@dataclass(unsafe_hash=True)
class Withdrawing(PlutusData):
    """
    For scripts that validate reward withdrawals from a reward account.
    The argument identifies the target reward account.
    """

    CONSTR_ID = 2
    staking_credential: StakingCredential


@dataclass(unsafe_hash=True)
class ConstitutionalCommitteeMember(PlutusData):
    CONSTR_ID = 0
    credential: Credential


@dataclass(unsafe_hash=True)
class DelegateRepresentative(PlutusData):
    CONSTR_ID = 1
    credential: Credential


@dataclass(unsafe_hash=True)
class StakePool(PlutusData):
    CONSTR_ID = 2
    verification_key_hash: PubKeyHash


Voter = Union[ConstitutionalCommitteeMember, DelegateRepresentative, StakePool]


@dataclass(unsafe_hash=True)
class GovernanceActionId(PlutusData):
    CONSTR_ID = 0
    transaction: TxId
    # Note: is a non-negative index to the proposal
    proposal_procedure: int


@dataclass(unsafe_hash=True)
class VoteNo(PlutusData):
    CONSTR_ID = 0


@dataclass(unsafe_hash=True)
class VoteYes(PlutusData):
    CONSTR_ID = 1


@dataclass(unsafe_hash=True)
class VoteAbstain(PlutusData):
    CONSTR_ID = 2


Vote = Union[VoteNo, VoteYes, VoteAbstain]


@dataclass(unsafe_hash=True)
class SomeGovernanceActionId(PlutusData):
    """
    Indicates that there is an governance action id associated with this output, given in the first field
    """

    CONSTR_ID = 0
    governance_action_id: GovernanceActionId


@dataclass(unsafe_hash=True)
class NoGovernanceActionId(PlutusData):
    """
    Indicates that there is no governance action id associated with this output
    """

    CONSTR_ID = 1


# To convert these values to interpretable values,
# use the index -> value mapping from aiken
# https://github.com/aiken-lang/stdlib/blob/2.2.0/lib/cardano/governance/protocol_parameters.ak
# Explicit mapping functions may follow
ProtocolParametersUpdate = Dict[int, Datum]

MaybeGovernanceActionId = Union[SomeGovernanceActionId, NoGovernanceActionId]

# Below are Governance Actions (GAXXX)


@dataclass(unsafe_hash=True)
class GAProtocolParameters(PlutusData):
    CONSTR_ID = 0
    # The last governance action of type 'ProtocolParameters'. They must all form a chain.
    ancestor: MaybeGovernanceActionId
    # The new proposed protocol parameters. Only values set to `Some` are relevant.
    new_parameters: ProtocolParametersUpdate
    # The optional guardrails script defined in the constitution.
    # The script is executed by the ledger in addition to the hard-coded ledger rules.
    # It must pass for the new protocol parameters to be deemed valid.
    guardrails: Union[SomeScriptHash, NoScriptHash]


@dataclass(unsafe_hash=True)
class ProtocolVersion(PlutusData):
    CONSTR_ID = 0
    major: int
    minor: int


@dataclass(unsafe_hash=True)
class GAHardFork(PlutusData):
    CONSTR_ID = 1
    # The last governance action of type 'HardFork'. They must all form a chain.
    ancestor: MaybeGovernanceActionId
    # The new proposed version. A few rules apply to proposing new versions:
    #
    # - The `major` component, if incremented, must be exactly one more than the current.
    # - The `minor` component, if incremented, must be exactly one more than the current.
    # - If the `major` component is incremented, `minor` must be set to `0`.
    # - Neither `minor` nor `major` can be decremented.
    new_version: ProtocolVersion


@dataclass(unsafe_hash=True)
class GATreasuryWithdrawal(PlutusData):
    CONSTR_ID = 2
    #  A collection of beneficiaries, which can be plain verification key
    #  hashes or script hashes (e.g. DAO).
    beneficiaries: Dict[Credential, Lovelace]
    # The optional guardrails script defined in the constitution. The script
    # is executed by the ledger in addition to the hard-coded ledger rules.
    #
    # It must pass for the withdrawals to be authorized.
    guardrails: SomeScriptHash


@dataclass(unsafe_hash=True)
class GANoConfidence(PlutusData):
    CONSTR_ID = 3
    # The last governance action of type `NoConfidence` or `ConstitutionalCommittee`.
    # They must all form a chain.
    ancestor: MaybeGovernanceActionId


@dataclass(unsafe_hash=True)
class GAConstitutionalCommittee(PlutusData):
    CONSTR_ID = 4
    # The last governance action of type `NoConfidence` or `ConstitutionalCommittee`.
    # They must all / form a chain.
    ancestor: MaybeGovernanceActionId
    # Constitutional members to be removed.
    evicted_members: List[Credential]
    # Constitutional members to be added.
    # Int value is a "mandate", an epoch number after which constitutional committee member
    # mandate expires.
    added_members: Dict[Credential, int]
    # The new quorum value, as a ratio of a numerator and a denominator.
    # The quorum specifies the threshold of 'Yes' votes necessary for the constitutional committee to accept a proposal procedure.
    quorum: Fraction


@dataclass(unsafe_hash=True)
class Constitution(PlutusData):
    CONSTR_ID = 0
    # Guardrails for operations as a script
    guardrails: MaybeScriptHash


@dataclass(unsafe_hash=True)
class GANewConstitution(PlutusData):
    CONSTR_ID = 5
    # The last governance action of type `Constitution` or `ConstitutionalCommittee`.
    # They must all form a chain.
    ancestor: MaybeGovernanceActionId
    constitution: Constitution


@dataclass(unsafe_hash=True)
class GAInfo(PlutusData):
    # No effect on chain, only informative nature
    CONSTR_ID = 6


GovernanceAction = Union[
    GAProtocolParameters,
    GAHardFork,
    GATreasuryWithdrawal,
    GANoConfidence,
    GAConstitutionalCommittee,
    GANewConstitution,
    GAInfo,
]


@dataclass(unsafe_hash=True)
class ProposalProcedure(PlutusData):
    """
    Represents a proposal procedure in governance.
    """

    deposit: Lovelace
    return_address: Credential
    governance_action: GovernanceAction


@dataclass(unsafe_hash=True)
class Publishing(PlutusData):
    """
    Needed when delegating to a pool using stake credentials defined as a
    custom script. This purpose is also triggered when de-registering such
    stake credentials.

    The index is a 0-based index of the given `Certificate` in `certificates`.
    """

    CONSTR_ID = 3
    certificate_index: int
    certificate: Certificate


@dataclass(unsafe_hash=True)
class Voting(PlutusData):
    """
    Voting for a type of voter using a governance action id to vote
    yes / no / abstain inside a transaction.
    The voter is who is doing the governance action.
    """

    CONSTR_ID = 4
    voter: Voter


@dataclass(unsafe_hash=True)
class Proposing(PlutusData):
    """
    Used to propose a governance action.
    A 0-based index of the given `ProposalProcedure` in `proposal_procedures`.
    """

    CONSTR_ID = 5
    proposal_index: int
    proposal_procedure: ProposalProcedure


# The reason that this script is being invoked
ScriptPurpose = Union[Minting, Spending, Withdrawing, Publishing, Voting, Proposing]


@dataclass(unsafe_hash=True)
class TxInfo(PlutusData):
    """
    A complex agglomeration of everything that could be of interest to the executed script, regarding the transaction
    that invoked the script
    """

    CONSTR_ID = 0
    inputs: List[TxInInfo]
    reference_inputs: List[TxInInfo]
    outputs: List[TxOut]
    fee: Lovelace
    mint: Value
    certificates: List[Certificate]
    # NOTE: Withdrawals are ordered by ascending Credential. Yet, note that `Script` credentials are treated as **lower values** than `VerificationKey` credentials.
    withdrawals: Dict[StakingCredential, int]
    validity_range: POSIXTimeRange
    # NOTE: Redeemers are ordered by ascending ScriptPurpose.
    signatories: List[PubKeyHash]
    redeemers: Dict[ScriptPurpose, Redeemer]
    datums: Dict[DatumHash, Datum]
    id: TxId
    # NOTE: Votes are ordered by ascending Voter and GovernanceActionId. First constructor variants in a type are treated as lower indices; except for Credential where `Script` credentials are treated as **lower values** than `VerificationKey` credentials.
    votes: Dict[Voter, Dict[GovernanceActionId, Vote]]
    proposal_procedures: List[ProposalProcedure]
    current_treasury_amount: OptionalLovelace
    treasury_donation: OptionalLovelace


@dataclass(unsafe_hash=True)
class ScriptContext(PlutusData):
    """
    Auxiliary information about the transaction and reason for invocation of the called script.
    """

    CONSTR_ID = 0
    transaction: TxInfo
    redeemer: Redeemer
    purpose: ScriptPurpose



import ast as _ast
import typing as _typing
import pluthon as _plt

from .type_impls import NoneInstanceType, Type


class TypedAST(_ast.AST):
    typ: "Type"


class typedexpr(TypedAST, _ast.expr):
    def typechecks(self) -> _typing.Dict[str, "Type"]:
        """Successful typechecks if this expression evaluates to True"""
        return {}


class typedstmt(TypedAST, _ast.stmt):
    # Statements always have type None
    typ = NoneInstanceType


class typedarg(TypedAST, _ast.arg):
    pass


class typedarguments(TypedAST, _ast.arguments):
    args: _typing.List[typedarg]
    vararg: _typing.Union[typedarg, None]
    kwonlyargs: _typing.List[typedarg]
    kw_defaults: _typing.List[_typing.Union[typedexpr, None]]
    kwarg: _typing.Union[typedarg, None]
    defaults: _typing.List[typedexpr]


class TypedModule(typedstmt, _ast.Module):
    body: _typing.List[typedstmt]


class TypedFunctionDef(typedstmt, _ast.FunctionDef):
    body: _typing.List[typedstmt]
    args: typedarguments
    orig_name: str


class TypedIf(typedstmt, _ast.If):
    test: typedexpr
    body: _typing.List[typedstmt]
    orelse: _typing.List[typedstmt]


class TypedReturn(typedstmt, _ast.Return):
    value: typedexpr


class TypedExpression(typedstmt, _ast.Expression):
    body: typedexpr


class TypedCall(typedexpr, _ast.Call):
    func: typedexpr
    args: _typing.List[typedexpr]


class TypedExpr(typedstmt, _ast.Expr):
    value: typedexpr


class TypedAssign(typedstmt, _ast.Assign):
    targets: _typing.List[typedexpr]
    value: typedexpr


class TypedClassDef(typedstmt, _ast.ClassDef):
    class_typ: "Type"


class TypedAnnAssign(typedstmt, _ast.AnnAssign):
    target: typedexpr
    annotation: "Type"
    value: typedexpr


class TypedWhile(typedstmt, _ast.While):
    test: typedexpr
    body: _typing.List[typedstmt]
    orelse: _typing.List[typedstmt]


class TypedFor(typedstmt, _ast.For):
    target: typedexpr
    iter: typedexpr
    body: _typing.List[typedstmt]
    orelse: _typing.List[typedstmt]


class TypedPass(typedstmt, _ast.Pass):
    pass


class TypedName(typedexpr, _ast.Name):
    pass


class Typedkeyword(TypedAST, _ast.keyword):
    arg: typedexpr
    value: typedexpr


class TypedConstant(TypedAST, _ast.Constant):
    pass


class TypedTuple(typedexpr, _ast.Tuple):
    pass


class TypedList(typedexpr, _ast.List):
    pass


class typedcomprehension(typedexpr, _ast.comprehension):
    target: typedexpr
    iter: typedexpr
    ifs: _typing.List[typedexpr]


class TypedListComp(typedexpr, _ast.ListComp):
    elt: typedexpr
    generators: _typing.List[typedcomprehension]


class TypedDictComp(typedexpr, _ast.DictComp):
    key: typedexpr
    value: typedexpr
    generators: _typing.List[typedcomprehension]


class TypedFormattedValue(typedexpr, _ast.FormattedValue):
    value: typedexpr
    conversion: int
    format_spec: _typing.Optional[_ast.JoinedStr]


class TypedJoinedStr(typedexpr, _ast.JoinedStr):
    values: _typing.List[typedexpr]


class TypedDict(typedexpr, _ast.Dict):
    pass


class TypedIfExp(typedstmt, _ast.IfExp):
    test: typedexpr
    body: typedexpr
    orelse: typedexpr


class TypedCompare(typedexpr, _ast.Compare):
    left: typedexpr
    ops: _typing.List[_ast.cmpop]
    comparators: _typing.List[typedexpr]


class TypedBinOp(typedexpr, _ast.BinOp):
    left: typedexpr
    right: typedexpr


class TypedBoolOp(typedexpr, _ast.BoolOp):
    values: _typing.List[typedexpr]


class TypedUnaryOp(typedexpr, _ast.UnaryOp):
    operand: typedexpr


class TypedSubscript(typedexpr, _ast.Subscript):
    value: typedexpr


class TypedAttribute(typedexpr, _ast.Attribute):
    value: typedexpr
    pos: int


class TypedAssert(typedstmt, _ast.Assert):
    test: typedexpr
    msg: typedexpr


class TypedSlice(typedexpr, _ast.Slice):
    lower: _typing.Optional[typedexpr]
    upper: _typing.Optional[typedexpr]
    step: _typing.Optional[typedexpr]


class RawPlutoExpr(typedexpr):
    typ: "Type"
    expr: _plt.AST

    _attributes = ["lineno", "col_offset", "end_lineno", "end_col_offset"]
    _fields = []



from ast import *

from ..util import CompilingNodeTransformer

"""
Make sure that certain variable names may never be overwritten
"""

FORBIDDEN_NAMES = {
    # type check
    "isinstance",
    # Type annotations
    "List",
    "Dict",
    "Union",
    "Self",
    # decorator and class name
    "dataclass",
    "PlutusData",
    # special decorator marking wrapped builtins
    "wraps_builtin",
}


class ForbiddenOverwriteError(ValueError):
    pass


class RewriteForbiddenOverwrites(CompilingNodeTransformer):
    step = "Checking for forbidden name overwrites"

    def visit_Name(self, node: Name) -> Name:
        if isinstance(node.ctx, Store) and node.id in FORBIDDEN_NAMES:
            raise ForbiddenOverwriteError(
                f"It is not allowed to overwrite name {node.id}"
            )
        return node

    def visit_FunctionDef(self, node: FunctionDef) -> FunctionDef:
        # Check if the function name is forbidden
        if node.name in FORBIDDEN_NAMES:
            raise ForbiddenOverwriteError(
                f"It is not allowed to overwrite name {node.name}"
            )
        for arg in node.args.args:
            if arg.arg in FORBIDDEN_NAMES:
                raise ForbiddenOverwriteError(
                    f"It is not allowed to overwrite name {arg.arg}"
                )
        node.body = [self.visit(n) for n in node.body]
        return node

    def visit_ClassDef(self, node: ClassDef) -> ClassDef:
        # Check if the class name is forbidden
        if node.name in FORBIDDEN_NAMES:
            raise ForbiddenOverwriteError(
                f"It is not allowed to overwrite name {node.name}"
            )
        node.body = [self.visit(n) for n in node.body]
        return node



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
.. include:: ../README.md
"""

import warnings
import importlib.metadata

__version__ = importlib.metadata.version(__package__ or __name__)
__author__ = "nielstron"
__author_email__ = "niels@opshin.dev"
__copyright__ = "Copyright (C) 2025 nielstron"
__license__ = "MIT"
__url__ = "https://github.com/OpShin/opshin"


try:
    from .compiler import *
    from .builder import *
    from .util import CompilerError
except ImportError as e:
    warnings.warn(ImportWarning(e))



from _ast import BoolOp, Call, FunctionDef, If, UnaryOp
from ast import *
from typing import Any, List
from ..util import CompilingNodeTransformer
from copy import deepcopy

"""
Expand union types
"""


def type_to_suffix(typ: expr) -> str:
    try:
        raw = unparse(typ)
    except Exception:
        return "UnknownType"
    return (
        raw.replace(" ", "")
        .replace("__", "___")
        .replace("[", "_l_")
        .replace("]", "_r_")
        .replace(",", "_c_")
        .replace(".", "_d_")
    )


class RemoveDeadCode(CompilingNodeTransformer):
    def __init__(self, arg_types: dict[str, str]):
        self.arg_types = arg_types

    def visit_FunctionDef(self, node: FunctionDef) -> Any:
        node.body = self.visit_sequence(node.body)
        return node

    def visit_sequence(self, stmts):
        new_stmts = []
        for stmt in stmts:
            s = self.visit(stmt)
            if isinstance(s, If) and isinstance(s.test, Constant):
                if s.test.value:
                    new_stmts.extend(s.body)
                else:
                    new_stmts.extend(s.orelse)
            else:
                new_stmts.append(s)
        return new_stmts

    def visit_If(self, node: If) -> Any:
        """
        Common types for `ast.If.test`:

            ast.Name      - `if x:`                     (truthiness of a variable)
            ast.Constant  - `if True:`, `if 0:`         (literal truthy/falsy)
            ast.Call      - `if func()`, `isinstance()` (function call)
            ast.Compare   - `if x > 3:`                 (comparison)
            ast.BoolOp    - `if x and y:`               (`and` / `or` logic)
            ast.UnaryOp   - `if not x:`                 (negation, e.g. `not`)
            ast.BinOp     - `if x + y:`                 (binary operation)
            ast.Attribute - `if obj.ready:`             (attribute access)
            ast.Subscript - `if arr[0]:`                (indexing)
            ast.Lambda    - `if lambda x: x > 0:`       (lambda - rare)
            ast.IfExp     - `if a if cond else b:`      (ternary - rare)

            The most likely to be used are ast.Call (if isinstance(...)), ast.BoolOp (if isinstance(...) and/or isinstance(...)), and ast.UnaryOp (if not isinstance(...))
        """
        node.test = self.visit(node.test)
        node.body = self.visit_sequence(node.body)
        node.orelse = self.visit_sequence(node.orelse)
        return node

    def visit_Call(self, node: Call) -> Any:
        node = self.generic_visit(node)
        # Check if this is an isinstance(x, T) call
        if (
            isinstance(node.func, Name)
            and node.func.id == "isinstance"
            and len(node.args) == 2
        ):
            arg, typ = node.args
            if isinstance(arg, Name) and isinstance(typ, Name):
                known_type = self.arg_types.get(arg.id)
                if known_type is not None:
                    typ_str = getattr(typ, "id", type_to_suffix(typ))
                    return Constant(value=(known_type == typ_str))

        return node

    def visit_BoolOp(self, node: BoolOp) -> Any:
        node.values = [self.visit(v) for v in node.values]
        # Check if all values are constants
        if all(isinstance(v, Constant) for v in node.values):
            values = [bool(v.value) for v in node.values]
            if isinstance(node.op, And):
                return Constant(value=all(values))
            elif isinstance(node.op, Or):
                return Constant(value=any(values))

        # Partial simplification: drop neutral constants
        # e.g. in `x or True`, return Constant(True)
        # e.g. in `x and False`, return Constant(False)
        if isinstance(node.op, And):
            for v in node.values:
                if isinstance(v, Constant) and not v.value:
                    return Constant(value=False)  # short-circuit
            node.values = [
                v for v in node.values if not (isinstance(v, Constant) and v.value)
            ]
        elif isinstance(node.op, Or):
            for v in node.values:
                if isinstance(v, Constant) and v.value:
                    return Constant(value=True)  # short-circuit
            node.values = [
                v for v in node.values if not (isinstance(v, Constant) and not v.value)
            ]
        # If only one value remains, return it directly
        if len(node.values) == 1:
            return node.values[0]
        return node

    def visit_UnaryOp(self, node: UnaryOp) -> Any:
        node.operand = self.visit(node.operand)

        # Only handle 'not' operations for now
        if isinstance(node.op, Not):
            # If it's `not <constant>`, simplify it
            if isinstance(node.operand, Constant):
                return Constant(value=not bool(node.operand.value))

        return node

    def visit_IfExp(self, node: IfExp) -> Any:
        node.test = self.visit(node.test)
        node.body = self.visit(node.body)
        node.orelse = self.visit(node.orelse)

        # Simplify if the test condition is a constant
        if isinstance(node.test, Constant):
            if node.test.value:
                return node.body
            else:
                return node.orelse

        return node


class OptimizeUnionExpansion(CompilingNodeTransformer):
    step = "Expanding Unions"

    def visit(self, node):
        if hasattr(node, "body") and isinstance(node.body, list):
            node.body = self.visit_sequence(node.body)
        if hasattr(node, "orelse") and isinstance(node.orelse, list):
            node.orelse = self.visit_sequence(node.orelse)
        if hasattr(node, "finalbody") and isinstance(node.finalbody, list):
            node.finalbody = self.visit_sequence(node.finalbody)
        return super().visit(node)

    def is_Union_annotation(self, ann: expr):
        if isinstance(ann, Subscript) and isinstance(ann.value, Name):
            if ann.value.id == "Union":
                return ann.slice.elts
        return False

    def split_functions(
        self, stmt: FunctionDef, args: list, arg_types: dict, naming=""
    ) -> List[FunctionDef]:
        """
                Recursively generate variants of a function with all possible combinations
        of expanded union types for its arguments.
        """
        new_functions = []
        for i, arg in enumerate(args):
            if not arg:
                continue
            n_args = deepcopy(args)
            n_args[i] = False
            for typ in arg:
                new_f = deepcopy(stmt)
                new_f.args.args[i].annotation = typ
                typ_str = getattr(typ, "id", type_to_suffix(typ))
                new_f.name = f"{naming}_{typ_str}"
                new_arg_types = deepcopy(arg_types)
                new_arg_types[stmt.args.args[i].arg] = typ_str
                new_f = RemoveDeadCode(new_arg_types).visit(new_f)
                new_functions.append(new_f)
                new_functions.extend(
                    self.split_functions(new_f, n_args, new_arg_types, new_f.name)
                )
            # Look for variation where this arg is still Union
            new_functions.extend(
                self.split_functions(stmt, n_args, arg_types, f"{naming}_Union")
            )
            # Handle only one Union per recursion level
            break

        return new_functions

    def visit_sequence(self, body):
        new_body = []
        for stmt in body:
            new_body.append(stmt)
            if isinstance(stmt, FunctionDef):
                args = [
                    self.is_Union_annotation(arg.annotation) for arg in stmt.args.args
                ]
                # number prefix here should guarantee naming uniqueness
                new_funcs = self.split_functions(stmt, args, {}, stmt.name + "+")
                # track variants
                new_body[-1].expanded_variants = [f.name for f in new_funcs]
                new_body.extend(new_funcs)
        return new_body



from ast import *
from copy import deepcopy

from ..util import CompilingNodeTransformer

"""
Rewrites all occurrences of comparison chaining into normal comparisons.
"""


class RewriteComparisonChaining(CompilingNodeTransformer):
    step = "Rewriting comparison chaining"

    def visit_Compare(self, node: Compare) -> Compare:
        if len(node.ops) <= 1:
            return self.generic_visit(node)
        all_comparators = [node.left] + node.comparators
        compare_values = []
        for comparator_left, comparator_right, op in zip(
            all_comparators[:-1], all_comparators[1:], node.ops
        ):
            # Deep copy comparators to avoid sharing AST nodes between comparisons
            # This prevents scoping issues when the same expression appears multiple times
            compare_values.append(
                Compare(
                    left=deepcopy(comparator_left),
                    ops=[op],
                    comparators=[deepcopy(comparator_right)],
                )
            )
        new_node = BoolOp(
            op=And(),
            values=compare_values,
        )
        return self.generic_visit(new_node)



from ast import *

from ..typed_ast import TypedCall, TypedAssert, TypedExpr
from ..fun_impls import PrintImpl
from ..type_impls import PolymorphicFunctionInstanceType
from ..util import CompilingNodeTransformer

"""
Removes traces in print and assert statements
"""


class OptimizeRemoveTrace(CompilingNodeTransformer):
    step = (
        "Removing occurrences of 'print' and 'assert', that use only constant strings"
    )

    def visit_Assert(self, node: TypedAssert):
        return TypedAssert(
            test=node.test,
            msg=None,
            lineno=node.lineno,
            col_offset=node.col_offset,
        )

    def visit_Expr(self, node: TypedExpr):
        if isinstance(node.value, Call):
            node_call = node.value
            if (
                isinstance(node_call.func.typ, PolymorphicFunctionInstanceType)
                and isinstance(node_call.func.typ.polymorphic_function, PrintImpl)
                and all(isinstance(arg, Constant) for arg in node_call.args)
            ):
                return None
        return node



"""
A std library that imports all valid hash functions from the builtin python hashlib library:
If you want to have all hash functions in scope, simply add `from opshin.std.hashlib import *` to the top of you python file.

Use it like this:
```python
from hashlib import sha256, sha3_256, blake2b

# Now you can use the hash functions directly
sha256_hash: bytes = sha256(b"Hello, World!").digest()
sha3_256_hash: bytes = sha3_256(b"Hello, World!").digest()
blake2b_hash: bytes = blake2b(b"Hello, World!").digest()
```
"""

from hashlib import sha256, sha3_256, blake2b
